# Copyright (c) 2019-present, The Johann Authors. All Rights Reserved.
# Use of this source code is governed by a BSD-3-clause license that can
# be found in the LICENSE file. See the AUTHORS file for names of contributors.

"""Johann, lightweight and flexible scenario orchestration"""

__version__ = "0.2.1-alpha"
